"""
"""

import asyncio
import unittest
import uuid
import aiohttp
from aiogremlin import (submit, GremlinConnector, GremlinClient,
                        GremlinClientSession, GremlinServerError,
                        GremlinClientWebSocketResponse)


class SubmitTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

    def tearDown(self):
        self.loop.close()

    def test_submit(self):

        @asyncio.coroutine
        def go():
            resp = yield from submit("x + x", url='https://localhost:8182/',
                                     bindings={"x": 4}, loop=self.loop,
                                     username="stephen", password="password")
            results = yield from resp.get()
            return results
        results = self.loop.run_until_complete(go())
        self.assertEqual(results[0].data[0], 8)

    def test_rebinding(self):

        @asyncio.coroutine
        def go1():
            result = yield from submit("graph2.addVertex()",
                                       url='https://localhost:8182/',
                                       loop=self.loop, username="stephen",
                                       password="password")
            resp = yield from result.get()

        try:
            self.loop.run_until_complete(go1())
            error = False
        except GremlinServerError:
            error = True
        self.assertTrue(error)

        @asyncio.coroutine
        def go2():
            result = yield from submit(
                "graph2.addVertex()", url='https://localhost:8182/',
                rebindings={"graph2": "graph"}, loop=self.loop,
                username="stephen", password="password")
            resp = yield from result.get()
            self.assertEqual(len(resp), 1)

        try:
            self.loop.run_until_complete(go2())
        except GremlinServerError:
            print("RELEASE DOES NOT SUPPORT REBINDINGS")


class GremlinClientTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        connector = aiohttp.TCPConnector(force_close=False, loop=self.loop,
                                         verify_ssl=False)

        client_session = aiohttp.ClientSession(
            connector=connector, loop=self.loop,
            ws_response_class=GremlinClientWebSocketResponse)

        self.gc = GremlinClient(url="https://localhost:8182/", loop=self.loop,
                                username="stephen", password="password",
                                client_session=client_session)

    def tearDown(self):
        self.loop.run_until_complete(self.gc.close())
        self.loop.close()

    def test_connection(self):

        @asyncio.coroutine
        def go():
            ws = yield from self.gc._connector.ws_connect(self.gc.url)
            self.assertFalse(ws.closed)
            yield from ws.close()

        self.loop.run_until_complete(go())

    def test_execute(self):

        @asyncio.coroutine
        def go():
            resp = yield from self.gc.execute("x + x", bindings={"x": 4})
            return resp

        results = self.loop.run_until_complete(go())
        self.assertEqual(results[0].data[0], 8)

    def test_sub_waitfor(self):
        sub1 = self.gc.execute("x + x", bindings={"x": 1})
        sub2 = self.gc.execute("x + x", bindings={"x": 2})
        sub3 = self.gc.execute("x + x", bindings={"x": 4})
        coro = asyncio.gather(*[asyncio.async(sub1, loop=self.loop),
                              asyncio.async(sub2, loop=self.loop),
                              asyncio.async(sub3, loop=self.loop)],
                              loop=self.loop)
        # Here I am looking for resource warnings.
        results = self.loop.run_until_complete(coro)
        self.assertIsNotNone(results)

    def test_resp_stream(self):
        @asyncio.coroutine
        def stream_coro():
            results = []
            resp = yield from self.gc.submit("x + x", bindings={"x": 4})
            while True:
                f = yield from resp.stream.read()
                if f is None:
                    break
                results.append(f)
            self.assertEqual(results[0].data[0], 8)
        self.loop.run_until_complete(stream_coro())

    def test_execute_error(self):
        execute = self.gc.execute("x + x g.asdfas", bindings={"x": 4})
        try:
            self.loop.run_until_complete(execute)
            error = False
        except:
            error = True
        self.assertTrue(error)

    def test_rebinding(self):
        execute = self.gc.execute("graph2.addVertex()")
        try:
            self.loop.run_until_complete(execute)
            error = False
        except GremlinServerError:
            error = True
        self.assertTrue(error)

        @asyncio.coroutine
        def go():
            result = yield from self.gc.execute(
                "graph2.addVertex()", rebindings={"graph2": "graph"})
            self.assertEqual(len(result), 1)

        try:
            self.loop.run_until_complete(go())
        except GremlinServerError:
            print("RELEASE DOES NOT SUPPORT REBINDINGS")


class GremlinClientSessionTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        connector = aiohttp.TCPConnector(force_close=False, loop=self.loop,
                                         verify_ssl=False)

        client_session = aiohttp.ClientSession(
            connector=connector, loop=self.loop,
            ws_response_class=GremlinClientWebSocketResponse)

        self.gc = GremlinClientSession(url="https://localhost:8182/",
                                       loop=self.loop,
                                       username="stephen", password="password",
                                       client_session=client_session)

        self.script1 = """v=graph.addVertex('name', 'Dave')"""

        self.script2 = "v.property('name')"

    def tearDown(self):
        self.loop.run_until_complete(self.gc.close())
        self.loop.close()

    def test_session(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            results = yield from self.gc.execute(self.script2)
            return results

        results = self.loop.run_until_complete(go())
        self.assertEqual(results[0].data[0]['value'], 'Dave')

    def test_session_reset(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            self.gc.reset_session()
            results = yield from self.gc.execute(self.script2)
            return results
        try:
            results = self.loop.run_until_complete(go())
            error = False
        except GremlinServerError:
            error = True
        self.assertTrue(error)

    def test_session_manual_reset(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            new_sess = str(uuid.uuid4())
            sess = self.gc.reset_session(session=new_sess)
            self.assertEqual(sess, new_sess)
            self.assertEqual(self.gc.session, new_sess)
            results = yield from self.gc.execute(self.script2)
            return results
        try:
            results = self.loop.run_until_complete(go())
            error = False
        except GremlinServerError:
            error = True
        self.assertTrue(error)

    def test_session_set(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            new_sess = str(uuid.uuid4())
            self.gc.session = new_sess
            self.assertEqual(self.gc.session, new_sess)
            results = yield from self.gc.execute(self.script2)
            return results
        try:
            results = self.loop.run_until_complete(go())
            error = False
        except GremlinServerError:
            error = True
        self.assertTrue(error)

    def test_resp_session(self):

        @asyncio.coroutine
        def go():
            session = str(uuid.uuid4())
            self.gc.session = session
            resp = yield from self.gc.submit("x + x", bindings={"x": 4})
            while True:
                f = yield from resp.stream.read()
                if f is None:
                    break
            self.assertEqual(resp.session, session)

        self.loop.run_until_complete(go())


if __name__ == "__main__":
    unittest.main()



"""Implements the Gremlin Server subprotocol."""

import base64
import collections
import uuid

try:
    import ujson as json
except ImportError:
    import json

from aiogremlin.exceptions import RequestError, GremlinServerError

__all__ = ("GremlinWriter", "gremlin_response_parser", "Message")


Message = collections.namedtuple(
    "Message",
    ["status_code", "data", "message", "metadata"])


def gremlin_response_parser(out, buf):
    while True:
        message = yield
        message = json.loads(message)
        message = Message(message["status"]["code"],
                          message["result"]["data"],
                          message["status"]["message"],
                          message["result"]["meta"])
        if message.status_code == 200:
            out.feed_data(message)
            out.feed_eof()
        elif message.status_code == 206 or message.status_code == 407:
            out.feed_data(message)
        elif message.status_code == 204:
            out.feed_data(message)
            out.feed_eof()
        else:
            if message.status_code < 500:
                raise RequestError(message.status_code, message.message)
            else:
                raise GremlinServerError(message.status_code, message.message)


class GremlinWriter:

    def __init__(self, ws):
        self.ws = ws

    def write(self, *, gremlin="", bindings=None, lang="gremlin-groovy",
              rebindings=None, op="eval", processor="", session=None,
              binary=True, mime_type="application/json", username="",
              password=""):
        if rebindings is None:
            rebindings = {}
        if op == "eval":
            message = self._prepare_message(gremlin,
                                            bindings,
                                            lang,
                                            rebindings,
                                            op,
                                            processor,
                                            session)

        if op == "authentication":
            message = self._authenticate(
                username, password, session, processor)
        message = json.dumps(message)
        if binary:
            message = self._set_message_header(message, mime_type)
        self.ws.send(message, binary=binary)
        return self.ws

    @staticmethod
    def _set_message_header(message, mime_type):
        if mime_type == "application/json":
            mime_len = b"\x10"
            mime_type = b"application/json"
        else:
            raise ValueError("Unknown mime type.")
        return b"".join([mime_len, mime_type, bytes(message, "utf-8")])

    @staticmethod
    def _prepare_message(gremlin, bindings, lang, rebindings, op, processor,
                         session):
        message = {
            "requestId": str(uuid.uuid4()),
            "op": op,
            "processor": processor,
            "args": {
                "gremlin": gremlin,
                "bindings": bindings,
                "language":  lang,
                "rebindings": rebindings
            }
        }
        if session is None:
            if processor == "session":
                raise RuntimeError("session processor requires a session id")
        else:
            message["args"].update({"session": session})
        return message


    @staticmethod
    def _authenticate(username, password, session, processor):
        auth = b"".join([b"\x00", bytes(username, "utf-8"), b"\x00", bytes(password, "utf-8")])
        message = {
            "requestId": str(uuid.uuid4()),
            "op": "authentication",
            "processor": processor,
            "args": {
                "sasl": base64.b64encode(auth).decode()
            }
        }
        if session is None:
            if processor == "session":
                raise RuntimeError("session processor requires a session id")
        else:
            message["args"].update({"session": session})
        return message



"""
Class used to pass messages with the Gremlin Server.
"""

import asyncio
import base64
import hashlib
import os

import aiohttp

from aiowebsocketclient.connector import ClientWebSocketResponse

__all__ = ('GremlinClientWebSocketResponse',)


class GremlinClientWebSocketResponse(ClientWebSocketResponse):
    """Wraps :py:class:`aiohttp.websocket_client.ClientWebSocketResponse`
    with minimal added functionality for the Gremln Server use case.
    """
    def __init__(self, reader, writer, protocol, response, timeout, autoclose,
                 autoping, loop):
        ClientWebSocketResponse.__init__(self, reader, writer, protocol,
                                         response, timeout, autoclose,
                                         autoping, loop)
        self._parser = aiohttp.StreamParser(buf=aiohttp.DataQueue(loop=loop),
                                            loop=loop)

    @property
    def parser(self):
        """
        Read-only property.

        :returns: :py:class:`aiohttp.parsers.StreamParser`
        """
        return self._parser

    @asyncio.coroutine
    def _close(self, *, code=1000, message=b''):
        if not self._closed:
            did_close = self._do_close()
            if did_close:
                return True
            while True:
                try:
                    msg = yield from asyncio.wait_for(
                        self._reader.read(), self._timeout, loop=self._loop)
                except asyncio.CancelledError:
                    self._close_code = 1006
                    self._response.close(force=True)
                    raise
                except Exception as exc:
                    self._close_code = 1006
                    self._exception = exc
                    self._response.close(force=True)
                    return True

                if msg.tp == aiohttp.MsgType.close:
                    self._close_code = msg.data
                    self._response.close(force=True)
                    return True
        else:
            return False

    def _do_close(self, code=1000, message=b''):
        self._closed = True
        try:
            self._writer.close(code, message)
        except asyncio.CancelledError:
            self._close_code = 1006
            self._response.close(force=True)
            raise
        except Exception as exc:
            self._close_code = 1006
            self._exception = exc
            self._response.close(force=True)
            return True

        if self._closing:
            self._response.close(force=True)
            return True

    def send(self, message, *, binary=True):
        """Send a message to the server."""
        if binary:
            method = self.send_bytes
        else:
            method = self.send_str
        try:
            method(message)
        except RuntimeError:
            # Socket closed.
            raise
        except TypeError:
            # Bytes/string input error.
            raise

    @asyncio.coroutine
    def receive(self):
        """
        :ref:`coroutine<coroutine>` method

        Receive a message from the server and push it into the parser.
        """
        msg = yield from super().receive()
        if msg.tp == aiohttp.MsgType.binary:
            self.parser.feed_data(msg.data.decode())
        elif msg.tp == aiohttp.MsgType.text:
            self.parser.feed_data(msg.data.strip())
        else:
            if msg.tp == aiohttp.MsgType.close:
                yield from ws.close()
            elif msg.tp == aiohttp.MsgType.error:
                raise msg.data
            elif msg.tp == aiohttp.MsgType.closed:
                pass



from setuptools import setup


setup(
    name="aiogremlin",
    version="0.1.3",
    url="",
    license="MIT",
    author="davebshow",
    author_email="davebshow@gmail.com",
    description="Python 3 driver for TP3 Gremlin Server built on Asyncio and aiohttp",
    long_description=open("README.txt").read(),
    packages=["aiogremlin", "tests"],
    install_requires=[
        "aiohttp==0.18.4",
        "aiowebsocketclient==0.0.5"
    ],
    test_suite="tests",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only'
    ]
)
